/*****************************************************************************
 *
 *  Parameter.h                                 XMM Science Analysis Subsystem
 *
 *  Parameter of a task
 *
 *  (c) 1998  ESA  -  Written mv Deby Jon Brumfitt, Aurora
 *
 *  $Id: Parameter.h,v 1.33 2001/12/19 16:57:24 jbakker Exp $
 *
 *****************************************************************************/

#ifndef PARAMETER_H
#define PARAMETER_H

#include <iostream>             // C++ headers
#include <string>
#include <list>
#include <vector>

#include "Optional.h"            // Application headers
#include "ParameterView.h"


//================================================================
//  Abstract Class Parameter
//================================================================

class Parameter
{
public:

    // Type definitions
    enum ParamType{Undefined, 
		   Bool, Int, Real, String, StringChoice, FileName,
		   BoolList, IntList, RealList, StringList, 
		   StringChoiceList, FileNameList,
		   Directory, DataSet, Block, Table, Array, Column, Plane,
		   DirectoryList, DataSetList, BlockList, TableList,
		   ArrayList, ColumnList, PlaneList,
                   Angle, Time, AngleList, TimeList,

		   // added types!
		   DirOrDataSet, DirOrFile
                  };

    // Constructors & destructors
    Parameter(const string& name = "", const string& prompt = "");
    virtual ~Parameter();

    // Virtual copy constructor
    virtual Parameter* clone() = 0;

    // View
    void setView(ParameterView* view);
    void addView(ParameterView* view);
    void deleteView(ParameterView* view=0);
    void notifyView();

    // Parameter name and prompt message
    const string& name()   const { return _name; }
    const string& prompt() const { return _prompt; }
    void setName(const string& name) { _name = name; }
    void setPrompt(const string& prompt) { _prompt = prompt; }


    // Accessing parameter attributes
    virtual ParamType   getType() const {return Undefined;}
    virtual const char* getTypeName() const { return ""; }
    virtual int         size() const {return -1;}

    // Accessing value
    virtual bool           boolValue() const;
    virtual int            intValue() const;
    virtual double         realValue() const;
    virtual const string&  stringValue() const;
    virtual vector<int>    boolListValue() const;
    virtual vector<int>    intListValue() const;
    virtual vector<double> realListValue() const;
    virtual vector<string> stringListValue() const;

    // Setting value
    virtual bool setNewValue(const string&) {return false;};
    virtual bool setValue(bool) { wrongType(); return false; }
    virtual bool setValue(int) { wrongType(); return false; }
    virtual bool setValue(double) { wrongType(); return false; }
    virtual bool setValue(const string&) { wrongType(); return false; }
    virtual bool setValue(const char*) { wrongType(); return false; }
//  Note: vector<bool> temporarily disabled to work-around egcs problem
//  virtual bool setValue(const vector<bool>&) { wrongType(); return false; }
    virtual bool setValue(const vector<int>&) { wrongType(); return false; }
    virtual bool setValue(const vector<double>&) { wrongType(); return false; }
    virtual bool setValue(const vector<string>&) { wrongType(); return false; }
    bool setValue(const Parameter*);

    // Defaults
    virtual bool setNewDefault(const string&) {return false;};
    virtual void revertToDefault() {};       // SHOULD BE "= 0"

    // Validation
    virtual bool validate() const {return true;}

    // Mandatory parameters
    bool  isSet() const { return _isSet; }
    bool  isMandatory() const { return _isMandatory; }
    void  setMandatory(bool mand) { _isMandatory = mand; }

    // Formatting
    virtual string asString() const = 0;
    virtual string defaultAsString() const { return _emptyString; }
    virtual void writeHelpToStream(ostream&) const;
    virtual void writeConstraint(ostream&) const {}
    friend ostream& operator<<(ostream& os, const Parameter& p);

protected:
    Parameter(const Parameter&);            // copy constructor is protected!
    Parameter& operator=(const Parameter&); // assignment (disabled)

    // Member data
    string         _name;
    string         _prompt;
    //    ParameterView* _view;

    // Mandatory parameters
    void setDefined() { _isSet = true; }

    virtual bool assignValueOf(const Parameter&)=0;

private:

    // Error handling
    void wrongType() const;

    // Mandatory parameters
    bool _isSet;
    bool _isMandatory;

    // All View available:
    class ParameterViews;
    ParameterViews* _views;

    // Constants
    static const string         _emptyString;
    static const vector<int>    _emptyBoolList;  // Temporary work-around for egcs
    static const vector<int>    _emptyIntList;
    static const vector<double> _emptyDoubleList;
    static const vector<string> _emptyStringList;
};

#endif

