/*****************************************************************************
 *
 *  StringListParameter.cc                      XMM Science Analysis Subsystem
 *
 *  List task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: StringListParameter.cc,v 1.15 2001/03/28 19:20:37 jbakker Exp $
 *
 *****************************************************************************/

#include <strstream>                // C++ headers
#include <iostream>
#include <string>
#include <vector>
#include <list>

#include "StringListParameter.h"      // Application headers
#include "StringToken.h"
#include <errstr.h>
#include <StringFns.h>
#include "freezestream.h"


//================================================================
//  Class StringListParameter
//================================================================

/*** Format the parameter value as a string in quotes ***/

string StringListParameter::asString() const
{
    ostrstream ostr;
    vector<string>::const_iterator it = _value.begin();
    ostr << '\'';
    if(it != _value.end())
        ostr << '"' << *it++ << '"';
    while(it != _value.end())
        ostr << " \"" << *it++ << '"';
    ostr << '\'' << ends;

    return freezestream(ostr);
}


/*** Format the parameter value as a string in quotes ***/

string StringListParameter::defaultAsString() const
{
    ostrstream ostr;
    vector<string>::const_iterator it = _default.begin();
    ostr << '\'';
    if(it != _default.end())
        ostr << '"' << *it++ << '"';
    while(it != _default.end())
        ostr << " \"" << *it++ << '"';
    ostr << '\'' << ends;

    return freezestream(ostr);
}


/*** Write a text representation of the constraint ***/

void StringListParameter::writeConstraint(ostream& os) const
{    
    if(!_choices.empty())
    {
        os << '{';
	list<string>::const_iterator it = _choices.begin();
	if(it != _choices.end())
	    os << *it++;
	while(it != _choices.end())
	    os << ' ' << *it++;
	os << '}';
    }
}


/***  Parse string representation and assign to 'value' if it is valid ***/

bool StringListParameter::parse(const string& str, vector<string>& value)
{
    list<string> tokens = StringFns::splitStringList(string(str));

    bool ok = true;
    list<string>::const_iterator it = tokens.begin();
    while(ok && it != tokens.end())
    {
        string s = *it++;
	StringToken t(s);
	string i = t.asString();
	if(t.isValid())
	    value.push_back(i);
	else {
	  errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");
	  ok = false;
	}
    }
    return ok;
}


/*** Revert to default value ***/

void StringListParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}


/*** Test whether the parameter is valid ***/

bool StringListParameter::validate() const
{
     return isValid(_value);
}


/*** Test whether value is allowable for the parameter ***/

bool StringListParameter::isValid(const vector<string>& value) const
{
    // Check that the values are in the list

    bool valid = true;
    if(!_choices.empty())
    {
        vector<string>::const_iterator it = value.begin();
	while(it != value.end())
	{
	    string s = *it++;

	    // This ought to use 'find', but it gives ambiguous "operator!="
	    // if(!(find(_choices.begin(),_choices.end(),s) == _choices.end()))

	    bool ok = false;
	    list<string>::const_iterator it = _choices.begin();
	    while(it != _choices.end() && !ok)
	    {
	        if(*it++ == s)
		  ok = true;
	    }
	    
	    if(!ok) 
	        valid = false;
	}
    }
    return valid;
}


/*** Set the parameter to the specified value ***/

bool StringListParameter::setValue(const vector<string>& value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }    
    return valid;
}


/*** Set the default to the specified value ***/

bool StringListParameter::setDefault(const vector<string>& value)
{
    bool valid = isValid(value);
    if(valid)
        _default = value;
    return valid;
}


/*** Set the value, specified as a string ***/

bool StringListParameter::setNewValue(const string& str)
{
    vector<string> values;
    bool ok =  parse(str, values) && setValue(values);

    if(!ok)
        reportRangeError(str);

    notifyView();
    return true;
}


/***  Set the default, specified as a string ***/

bool StringListParameter::setNewDefault(const string& str)
{    
    vector<string> values;
    bool ok = parse(str, values) && setDefault(values);

    if(!ok)
        reportRangeError(str);

    return true;
}


/*** Return the length of the list ***/

int StringListParameter::size() const
{
    return _value.size();
}


/*** Split a string into tokens ***/

list<string> StringListParameter::splitStringList(const string& str)
{
    list<string> strlist;
    const char* p = str.c_str();
    char c =*p++;

    while(c)
    {
        string token = "";

        // Skip leading space
	while(c && (c ==' ' || c == '\t')) 
	    c=*p++;

	// Quotes
	if(c == '"' || c == '\'') {
	    char quote = c;
	    c=*p++;
	    while(c && (c != quote)) {
	        token += c;
		c=*p++;
	    }
	    if(c != quote)
	      errstr << "Mismatched quotes for parameter '"
		     << _name << "' at token["<< strlist.size()+1 << "]="
		     << token
		     << Sas::error("ParamWrongQuotes");
	    c=*p++;
	    strlist.push_back(token);
	} 
	else if(c && (c != ' ' && c != '\t')) 
	{
	    // Simple token
	    while(c && (c != ' ' && c != '\t')) {
	        token += c;
		c=*p++;
	    }
	    strlist.push_back(token);
	}
    } 
    return strlist;
}


/*** Report that value is out of range ***/

void StringListParameter::reportRangeError(const string& valueString)
{
  errstr << "Parameter '" << _name << "';" << endl
	 << "value '" << valueString << "' is outside range ";
  writeConstraint(errstr);
  errstr << Sas::error("ParamNotInRange");
}
