/*****************************************************************************
 *
 *  XpiInfoReader.cc                            XMM Science Analysis Subsystem
 *
 *  Reader for XPI files
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: XpiInfoReader.cc,v 1.20.2.1 2002/10/29 12:47:21 gvacanti Exp $
 *
 *****************************************************************************/

#include "XpiInfoReader.h"     // Application headers
#include "Parameter.h"      
#include "BoolParameter.h"      
#include "IntParameter.h"      
#include "RealParameter.h" 
#include "AngleParameter.h"
#include "TimeParameter.h"     
#include "StringParameter.h"
#include "StringChoiceParameter.h"
#include "SpecialStringParameters.h"
#include "StringParameter.h"
#include "BoolListParameter.h"
#include "IntListParameter.h"
#include "RealListParameter.h"
#include "AngleListParameter.h"
#include "TimeListParameter.h"
#include "StringListParameter.h"
#include "StringChoiceListParameter.h"
#include "SpecialStringListParameters.h"
#include <errstr.h>
#include <SasFileLocator.h>
#include <StringFns.h>
#include "StringToken.h"

#include <values.h>            // C headers
#include <climits>

#include <iostream>          // C++ headers
#include <fstream>
#include <string>
#include <strstream>
#include <list>
#include <vector>


//================================================================
//  Class XPIParser
//================================================================

class XPIParser
{
public:
    XPIParser(const string& message="") : _err(false), _message(message) {}
    ParameterInfo* parse(istream& is);
    bool isValid() const { return !_err; }

private:
    StringToken  getToken();

    istrstream*    _ss;
    ParameterInfo* _info;
    bool           _err;
    string         _message;
};


/*** Get next token as a string ***/

StringToken XPIParser::getToken()
{
    const char quote ='"';

    string tok;
    char c = _ss->get();
    string token = "";

    // Skip leading space
    while(c == ' ' || c == '\t')
        c = _ss->get();

    while(c != EOF && c != ',' && c != '#') 
    {
        // Quotes - Copy everything up to closing quote
        if(c == quote) 
	{
	    token += quote;
	    c = _ss->get();
	    while(c != EOF && (c!=quote)) {
	        token += c;
		c = _ss->get();
	    }
	    if(c == EOF)
	      errstr << "Mismatched quotes reading token(" << token << ") in parameter file: "<< _message << Sas::error("ParamWrongQuotes");
	    token += quote;
	    c = _ss->get();
	    
	// Just copy anything else
	} else {
	    token += c;
	    c = _ss->get();
	}
    }

    // Remove trailing spaces
    // ("find_last_not_of" does not work with gcc-2.7.2.3)

    string::size_type n = token.length();
    while(n-- > 0 && (token[n] == ' ' || token[n] == '\t')) ;
    tok = token.substr(0,n+1);
    
    // Stop if comment found
    if(c == '#')
        _ss->putback(EOF);

    return StringToken(tok);
}


/*** Return a new ParameterInfo; caller must delete ***/

ParameterInfo* XPIParser::parse(istream& is)
{
    string buff;
    _info = new ParameterInfo;
    int lineno = 0;
    while(is)
    {
        bool err = false;
	string errmess;

	getline(is,buff);
	lineno++;
	_ss = new istrstream(buff.c_str());

	string name = getToken().asUnquotedString();
	if(name != "") {
	    string type      = getToken().asString();
	    string mode      = getToken().asUnquotedString();
	    StringToken def  = getToken();
	    StringToken min  = getToken();
	    StringToken max  = getToken();
	    string prompt    = getToken().asUnquotedString();


	    if ( !mode.empty() && !(mode == "h" || mode == "q") ) {
	        errmess += "unrecognised mode \"" + mode + "\"\n";
		err = true;
	    }
/*
	    cout << name
		 << '|' << type
		 << '|' << mode
		 << '|' << def
		 << '|' << min
		 << '|' << max
		 << '|' << prompt
		 << endl;
*/
	    Parameter *p = 0;
            char t = (type.length() > 0) ? type[0] : ' ';
            switch (t) {
          
	    // Boolean
	    case 'b': {
                string s = def.asString();

		if(!def.isValid()) 
		    err = true;
                p = new BoolParameter();
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
                break;
            }

            // Integer
	    case 'i': {
	        string s = def.asString();
		int minVal = min.asInt();  // Convert before testing isValid()
		int maxVal = max.asInt();
		if(!min.isValid() || !max.isValid())
		    err = true;
	        IntParameter* ip = new IntParameter();
		ip->setName(name);
		ip->setPrompt(prompt);
		ip->setNewDefault(s);
		ip->revertToDefault();

		if(!min.isUndef())
		    ip->setMin(minVal);
		if(!max.isUndef())
		    ip->setMax(maxVal);

		p = ip;
	        break;
	    }

            // Real
	    case 'r':
	    case 'u': {
	        string s = def.asString();
		double minVal = min.asDouble();  // Convert before testing isValid()
		double maxVal = max.asDouble();
		if(!min.isValid() || !max.isValid())
		    err = true;
		RealParameter* rp = 0;
		  switch (t) {
		    case 'r':
		        rp = new RealParameter();
		        break;
		    case 'u':
		        rp = new AngleParameter();
		        break;
		}
		rp->setName(name);
		rp->setPrompt(prompt);
		rp->setNewDefault(s);
		rp->revertToDefault();

		if(!min.isUndef())
		    rp->setMin(minVal);
		if(!max.isUndef())
		    rp->setMax(maxVal);

		p = rp;
                break;
	    }

            // Time
	    case 'w': {
	        string s = def.asUnquotedString();
		p = new TimeParameter();

		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }

            // String
	    case 's': {
		string s = def.asUnquotedString();
		if(!def.isValid())
		    err = true;
		string minVal = min.asUnquotedString();
		if(minVal == "")
		    p = new StringParameter();
		else
		{
		    StringChoiceParameter* ip = new StringChoiceParameter();
		    ip->setChoices(StringFns::splitDelimitedString(minVal,"|"));
		    p = ip;
		}
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }

            // Special string parameters
	    case 'y':
	    case 'e':
	    case 'z':
	    case 'x':
	    case 'f':
	    case 'k':
	    case 't':
	    case 'a':
	    case 'c':
	    case 'p':
    {
		string s = def.asUnquotedString();
		if(!def.isValid())
		    err = true;
//		string minVal = min.asString();

		switch (t) {
		  case 'y':
		    p = new DirectoryParameter();
		    break;
		  case 'e':
		    p = new DataSetParameter();
		    break;
		  case 'z':
		    p = new DirOrDataSetParameter();
		    break;
		  case 'x':
		    p = new DirOrFileParameter();
		    break;
		  case 'f':
		    p = new FileNameParameter();
		    break;
		  case 'k':
		    p = new BlockParameter();
		    break;
		  case 't':
		    p = new TableParameter();
		    break;
		  case 'a':
		    p = new ArrayParameter();
		    break;
		  case 'c':
		    p = new ColumnParameter();
		    break;
		  case 'p':
		    p = new PlaneParameter();
		    break;
		}
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }
             
	    // List of booleans
	    case 'B': {
	        string s = def.asString();
		vector<int> val;                   // Dummy
		
	        p = new BoolListParameter();
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }

	    // List of integers
	    case 'I': {
	        string s = def.asString();
		vector<int> val;                   // Dummy
		
		IntListParameter* ip = new IntListParameter();
		ip->setName(name);
		ip->setPrompt(prompt);
		ip->setNewDefault(s);
		ip->revertToDefault();
		
		if(!min.isUndef())
		    ip->setMin(min.asInt());
		if(!max.isUndef())
		    ip->setMax(max.asInt());

		p = ip;
	        break;
	    }

	    // List of reals
	    case 'R': 
	    case 'U':
	    case 'W': {
	        string s = def.asString();
		vector<double> val;                   // Dummy
		
		RealListParameter* rp = 0;
		  switch (t) {
		    case 'R':
		        rp = new RealListParameter();
		        break;
		    case 'W':
		        rp = new TimeListParameter();
		        break;
		    case 'U':
		        rp = new AngleListParameter();
		        break;
		}
		rp->setName(name);
		rp->setPrompt(prompt);
		rp->setNewDefault(s);
		rp->revertToDefault();
		
		if(!min.isUndef())
		    rp->setMin(min.asDouble());
		if(!max.isUndef())
		    rp->setMax(max.asDouble());

		p = rp;
	        break;
	    }

	    // List of strings
	    case 'S': {
	        string s = def.asString();
		string minVal = min.asUnquotedString();

		vector<string> val;                   // Dummy
	        if(minVal == "")
		    p = new StringListParameter();
		else
		{
		    StringChoiceListParameter* ip = new StringChoiceListParameter();
		    ip->setChoices(StringFns::splitDelimitedString(minVal,"|"));
		    p = ip;
		}
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }

	    // List of special strings
	    case 'Y':
	    case 'E':
	    case 'F':
	    case 'K':
	    case 'T':
	    case 'A':
	    case 'C':
	    case 'P':
	      {
	        string s = def.asString();
		vector<string> val;                   // Dummy
		
		switch (t) {
		  case 'Y':
		    p = new DirectoryListParameter();
		    break;
		  case 'E':
		    p = new DataSetListParameter();
		    break;
		  case 'F':
		    p = new FileNameListParameter();
		    break;
		  case 'K':
		    p = new BlockListParameter();
		    break;
		  case 'T':
		    p = new TableListParameter();
		    break;
		  case 'A':
		    p = new ArrayListParameter();
		    break;
		  case 'C':
		    p = new ColumnListParameter();
		    break;
		  case 'P':
		    p = new PlaneListParameter();
		    break;
		}
		p->setName(name);
		p->setPrompt(prompt);
		p->setNewDefault(s);
		p->revertToDefault();
	        break;
	    }

	    default:
	        errmess += "unrecognised type \"" + string(1,t) + "\"\n";
		err = true;
	        break;
            
	    }

	    if(p) {
	        if(mode == "q")
		    p->setMandatory(true);

		/*
		 * Should add check that parameter is not already defined
		 */

	        err |= !p->validate();
		_info->addParameter(p);
	    }
	}
    
	if(err)
	{
	    _err = true;
	    errstr << "Invalid definition of parameter \"" << name << "\" " << endl
		   << errmess
		   << "at line " << lineno << endl;
	    if(_message != "")
	        errstr << "in " << _message;
	    errstr << Sas::warning("invalidParam");
	}
    }
    delete _ss;
    _ss = 0;

    return _err ? 0 : _info;      // Null pointer if error
}
   

//================================================================
//  Class XpiInfoReader
//================================================================


XpiInfoReader::XpiInfoReader(const string& name)
{
    _info = 0;

    // Read the parameter file
    SasFileLocator locator("");
    string parFile = locator.locate("config/" + name + ".par", "", false);

    errstr << "Using parameter file " <<  parFile
	   << Sas::message(Msg::Lib, Msg::Noisy);

    ifstream is(parFile.c_str());
    
    if(is)
    {
         XPIParser p(parFile);
	 _info = p.parse(is);
	 if(_info)
	     _info->setTaskName(name);

    } else {
        errstr << "Could not open parameter file " << endl 
	       << parFile 
	       << Sas::error("ParamFileOpen");
    }
}
    

/*** Return a new ParameterInfo; caller must delete ***/

ParameterInfo* XpiInfoReader::getParamInfo()
{
    return _info;   // Null pointer if error
}

